#pragma once

#include <pthread.h>
#include <common/common.h>

enum {
    _LKT_PLAY_STOP   = 0,
    _LKT_PLAY_PLAY   = 1,
    _LKT_PLAY_PAUSE  = 2,
    _LKT_PLAY_TOGGLE = 3,
};

#define LKT_PLAY_STOP   ((void *) (size_t) _LKT_PLAY_STOP)
#define LKT_PLAY_PLAY   ((void *) (size_t) _LKT_PLAY_PLAY)
#define LKT_PLAY_PAUSE  ((void *) (size_t) _LKT_PLAY_PAUSE)
#define LKT_PLAY_TOGGLE ((void *) (size_t) _LKT_PLAY_TOGGLE)

enum lkt_event_type {
    lkt_event_null          = 0,        // NULL
    lkt_event_play_pos      = (1 << 1), // size_t
    lkt_event_play_file     = (1 << 2), // XXX: UNUSED //
    lkt_event_play_next     = (1 << 3), // NULL
    lkt_event_play_prev     = (1 << 4), // NULL
    lkt_event_play_toggle   = (1 << 5), // size_t, `LKT_PLAY_.*`
    lkt_event_prop_vol      = (1 << 6), // size_t
    lkt_event_prop_dur      = (1 << 7), // size_t
    lkt_event_prop_time     = (1 << 8), // size_t
    lkt_event_skip_current  = (1 << 9), // NULL
};

#define lkt_event_play ( lkt_event_play_pos | lkt_event_play_file |     \
                         lkt_event_play_next | lkt_event_play_prev |    \
                         lkt_event_play_toggle | lkt_event_skip_current \
                       )
#define lkt_event_prop ( lkt_event_prop_vol | lkt_event_prop_dur |      \
                         lkt_event_prop_time                            \
                       )

typedef struct {
    unsigned int type;
    void *attr;
} lkt_event;

struct queue {
    pthread_mutex_t lock;

    volatile lkt_event *contents;
    volatile size_t size;
    volatile size_t last;

    volatile int available;
};

int lkt_queue_new(struct queue *);
void lkt_queue_free(struct queue *);

void lkt_queue_send(struct queue *, enum lkt_event_type, void *attr);
lkt_event lkt_queue_handle(struct queue *);

void lkt_queue_make_available(struct queue *, enum lkt_event_type);
