#if ! defined(__REG_H__)
#define __REG_H__

struct lkt_state;

#include <common/macro.h>
#include <stdarg.h>

typedef int (*reg_func)(va_list *);

/* Store functions with a name, so that all the lib will have acces to it
   with a name that can be generated or red from a config file. */
struct module_reg {
    const char *name;
    union {
        reg_func func;
        struct module_reg **obj; /* First item of the array */
    };
};

/* A module */
struct lkt_module {
    struct module_reg *reg;
    void *handle;
    void *data;
};

#define REG_EXPORT_NAME     "__reg__"
#define REG_BEGIN(reg)      struct module_reg reg[] = {
#define REG_ADD(f)          { .name = #f,   .func = f    },
#define REG_ADD_NAMED(n, f) { .name = n,    .func = f    },
#define REG_REGISTER(n, s)  { .name = n,    .obj  = &s   },
#define REG_END()           { .name = NULL, .func = NULL } };
#define REG_DEFINE(reg)     struct module_reg reg* = NULL;

/* Use only once per .so files or you will have some problems... */
#if ! defined(_STATIC_MODULES)
#define REG_EXPORT(__reg)   struct module_reg *__reg__ = __reg;
#else
#define REG_EXPORT(__reg)
#endif

/* Import a reg */
#define REG_DECLARE(reg)    extern struct module_reg *reg;
#define REG_IMPORT(reg, as)             \
    REG_DECLARE(reg)                    \
    static struct module_reg *as = reg;

/* Call a function from a reg. */
#define MOD_CALL(module, name, ...)     \
    reg_call((module).reg, name, 1 + (GET_VA_COUNT(__VA_ARGS__)), &(module).data, __VA_ARGS__)
int reg_call(struct module_reg *reg, const char *name, int count, ...);

/* Import a module, from a file (mod is a filepath) or from the __reg__ in the
   current file. If the `mod` contains one of the following "/.\'\"\\", it is
   treated as a file. The function will first try to treat is as a module in the
   __reg__.
   Returns 0 on success, something else on error. */
int reg_import(const char *mod, struct module_reg **ret, void **handle);

/* Set the __reg__ pointer. */
void reg_export(struct module_reg *reg);

#endif /* __REG_H__ */
