#define _POSIX_C_SOURCE 200809L

#include <errno.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <common/common.h>
#include <common/queue.h>
#include <pthread.h>

int
lkt_queue_new(struct queue *ret)
{
    if (!ret)
        return 1;

    pthread_mutex_t mxt = PTHREAD_MUTEX_INITIALIZER;
    struct queue _ret = {
        .contents = malloc(LKT_DEFAULT_LIST_SIZE * sizeof(lkt_event)),
        .size = LKT_DEFAULT_LIST_SIZE,
        .last = 0,
        .available = 0,
        .lock = mxt,
    };

    if (_ret.contents == NULL)
        return 1;

    *ret = _ret;
    return 0;
}

void
lkt_queue_free(struct queue *queue)
{
    pthread_mutex_lock(&queue->lock);
    if (queue && queue->contents)
        free((void *) queue->contents);
    pthread_mutex_unlock(&queue->lock);
}

void
lkt_queue_send(struct queue *queue, enum lkt_event_type _type, void *_attr)
{
    pthread_mutex_lock(&queue->lock);
    GOTO_UNLESS(queue, "Invalid argument", end)

    volatile lkt_event *new;
    if (queue->size == queue->last) {
        new = realloc((void *) queue->contents, queue->size * 2 * sizeof(lkt_event));

        if (NULL == new)
            goto end;

        queue->contents = new;
        queue->size *= 2;
    }

    lkt_event evt = {
        .type = _type,
        .attr = _attr,
    };
    queue->contents[(queue->last)++] = evt;

    if (! (_type & queue->available))
        LOG_WARN("QUEUE", "The event %d is not available, push it anyway", _type);
end:
    pthread_mutex_unlock(&queue->lock);
}

lkt_event
lkt_queue_handle(struct queue *queue)
{
    pthread_mutex_lock(&queue->lock);
    lkt_event ret = {0};
    if (!queue || !queue->last)
        goto end;
    if (! (queue->contents[0].type & queue->available) &&
        queue->contents[0].type) {
        LOG_WARN("QUEUE", "Event %d is not available", queue->contents[0].type);
        goto end;
    }

    ret = queue->contents[0];
    memmove((void *) queue->contents, (void *) (queue->contents + 1),
            --(queue->last));
end:
    pthread_mutex_unlock(&queue->lock);
    return ret;
}

void
lkt_queue_make_available(struct queue *queue, enum lkt_event_type type)
{
    pthread_mutex_lock(&queue->lock);
    queue->available |= type;
    pthread_mutex_unlock(&queue->lock);
}
